/**
* A single mesh of a circuit
*/
function CircuitMesh(
    start       // CircuitPart instance that represents first item of mesh or CircuitMesh instance to clone
) {
    if (start instanceof CircuitMesh) {
        this._parts = start._parts.slice();
        this._gridpoints = start._gridpoints.slice();
        if (start.completed)
            this.completed = true;
    } else {
        this._parts = [ new CircuitMeshPart(start) ];
        this._gridpoints = start.getAllConnectors();
        if (start.isFlipped)
            this._gridpoints.reverse();
    }
}

/** All CircuitMeshPart instances in mesh order */
CircuitMesh.prototype._parts = [];

/** All grid points (e. g. { x: 2, y: 5 }) in mesh order */
CircuitMesh.prototype._gridpoints = [];

/** Is it a complete mesh reaching the secondary connection of the mesh's first component? */
CircuitMesh.prototype.completed = false;

/**
* Get grid point from which the mesh is to be continued
* @return Object        Grid point (e. g. { x: 2, y: 5 }) that is the open end of the incomplete mesh
*/
CircuitMesh.prototype.getOpenEnd = function() {
    if (this.completed)
        return null;
    return this._gridpoints[this._gridpoints.length - 1];
}

/**
* Create another CircuitMesh instance containing the same data
* @return CircuitMesh       Created clone instance
*/
CircuitMesh.prototype.clone = function() {
    return new CircuitMesh(this);
}

/**
* Checks if mesh (already) contains a CircuitPart instance
* @return Boolean       Does mesh contain needle?
*/
CircuitMesh.prototype.containsPart = function(
    needle          // CircuitPart instance to search for
) {
    return this._parts.filter(part => (part.part == needle)).length ? true : false;
}

/**
* Checks if mesh (already) contains a grid point
* @return Boolean       Does mesh contain gridpoint?
*/
CircuitMesh.prototype.containsGridPoint = function(
    gridpoint,      // Grid point (e. g. { x: 2, y: 5 }) to search for
    ignoreTarget    // Optional: Return false if gridpoint describes secondary connector of start component?
) {
    const points = ignoreTarget ? this._gridpoints.slice(1) : this._gridpoints;
    return points.filter(({ x, y }) => ((x == gridpoint.x) && (y == gridpoint.y))).length ? true : false;
}

/**
* Check if grid point closes the mesh
* @return Boolean       Does gridpoint complete the mesh?
*/
CircuitMesh.prototype.isTargetGridPoint = function(
    gridpoint       // Grid point (e. g. { x: 2, y: 5 }) to check
) {
    return ((this._gridpoints[0].x == gridpoint.x) && (this._gridpoints[0].y == gridpoint.y));
}

/**
* Adds a grid point to the list of contained grid nodes or mark this mesh as complete if gridpoint completes it
* @return undefined
*/
CircuitMesh.prototype._addGridPoint = function(
    gridpoint       // Grid point (e. g. { x: 2, y: 5 }) to add
) {
    if (this.isTargetGridPoint(gridpoint))
        this.completed = true;
    else
        this._gridpoints.push(gridpoint);
}

/**
* Add another part to the mesh (if possible)
* @return Boolean       Has part been added?
*/
CircuitMesh.prototype.add = function(
    circuitpart         // CircuitPart instance that represents this component to add within the circuit
) {
    const openend = this.getOpenEnd();
    if (!openend || (circuitpart == this._parts[0].part))
        return false;
    const gridpoints = circuitpart.getAllConnectors();
    switch (true) {
        case ((openend.x == gridpoints[0].x) && (openend.y == gridpoints[0].y) && !this.containsGridPoint(gridpoints[1], true)):
            this._parts.push(new CircuitMeshPart(circuitpart, false));
            this._addGridPoint(gridpoints[1]);
            return true;
        case ((openend.x == gridpoints[1].x) && (openend.y == gridpoints[1].y)&& !this.containsGridPoint(gridpoints[0], true)):
            this._parts.push(new CircuitMeshPart(circuitpart, true));
            this._addGridPoint(gridpoints[0]);
            return true;
        default:
            return false;
    }
}

/**
* Check if this mesh is a short circuit
* @return Boolean       Short circuit?
*/
CircuitMesh.prototype.isShortCircuit = function(
    requireVoltage          // Only report true if constraint voltage sums up to non-zero?
) {
    if (!this.completed)
        return false;
    if (this._parts.filter(part => part.component.getConstraintResistance() ? !part.component.getConstraintVoltage() : false).length)
        return false;
    if (requireVoltage && !this._parts.map(part => ((part.isSecondaryFirst ? (-1) : 1) * part.component.getConstraintVoltage())).reduce((a, b) => (a + b)))
        return false;
    return true;
}

/**
* Get all CircuitPart instances beeing part of this mesh
* @return Array     CircuitPart instances of this mesh
*/
CircuitMesh.prototype.getParts = function() {
    return this._parts.map(item => item.part);
}
