/**
* Render editable circuit on screen and manage editing
*/
function ControllerCircuitDisplay(
    containerNode       // DOM node to contain editable circuit
) {
    Object.defineProperty(this, 'circuit', { enumerable: true, writeable: false, value: new ModelCircuit() });
    this.node = containerNode;
    this.node = this.createSvg();
    containerNode.appendChild(this.node);
    var self = this;
    this.circuit.eventManager.addEventListener(true, function() { self.updateSvg(); });
    this.circuit.eventManager.addEventListener('restored', () => this._rearrangeCircuit() );
    this.node.ownerDocument.defaultView.addEventListener('resize', function() { self.updateSvg(); });
    // Start observer
    if (typeof ResizeObserver == 'function')
        (new ResizeObserver(function() {
            self._measureCm();
            self.updateSvg();
        })).observe(containerNode, { attributes: true });
}

/** Circuit to render */
ControllerCircuitDisplay.prototype.circuit = null;

/** DOM node showing circuit */
ControllerCircuitDisplay.prototype.node = null;

/** Current editing mode */
ControllerCircuitDisplay.prototype._editingMode = EditingModes[0];
Object.defineProperty(ControllerCircuitDisplay.prototype, 'editingMode', { enumerable: true,
    get() { return this._editingMode; },
    set(mode) {
        if (this._editingMode == mode)
            return;
        if (typeof this._editingMode.procStop == 'function')
            this._editingMode.procStop.call(this, this.circuit);
        this._editingMode = mode;
        if (typeof this._editingMode.procStart == 'function')
            this._editingMode.procStart.call(this, this.circuit);
    }
});

/** Local units per grid unit */
ControllerCircuitDisplay.prototype._scale = 20;
Object.defineProperty(ControllerCircuitDisplay.prototype, 'scale', { enumerable: true,
    get() { return this._scale; },
    set(v) {
        if (!Number.isFinite(v) || (v <= 0))
            return;
        this._scale = v;
        this.updateSvg();
    }
});

/** Horizontal offset of circuit origin */
ControllerCircuitDisplay.prototype._offsetX = undefined;

/** Vertical offset of circuit origin */
ControllerCircuitDisplay.prototype._offsetY = undefined;

/**
* Center circuit
* @return undefined
*/
ControllerCircuitDisplay.prototype.resetOffset = function() {
    this._offsetX = undefined;
    this._offsetY = undefined;
}

/**
* Explore (again) length of a centimetre in pixels and save value as this._cm
* @return undefined
*/
ControllerCircuitDisplay.prototype._measureCm = function() {
    const testnode = document.createElement('div');
    testnode.style = 'width:10mm';
    this.node.parentNode.appendChild(testnode);
    this._cm = testnode.clientWidth;
    this.node.parentNode.removeChild(testnode);
}

/**
* Rearrange circuit to fit in editing area
* @return Object        { x: "free space left of circuit (in grid units)", y: "free space above circuit (in grid units)" }
*/
ControllerCircuitDisplay.prototype._rearrangeCircuit = function(
    width,          // Available width in grid units (natural number!)
    height          // Available height in grid units (natural number!)
) {
    const used = this.circuit.getUsedGridArea();
    if (used.left || used.right || used.top || used.bottom) {
        this._offsetX = Math.floor((Number.isFinite(this._offsetX) ? Math.max(Math.min(this._offsetX, width - used.right), -used.left) : (width - used.left - used.right) / 2));
        this._offsetY = Math.floor((Number.isFinite(this._offsetY) ? Math.max(Math.min(this._offsetY, height - used.bottom), -used.top) : (height - used.top - used.bottom) / 2));
    } else {
        this._offsetX = 0;
        this._offsetY = 0;
    }
    return { x: this._offsetX + used.left, y: this._offsetY + used.top };
}

/**
* Build a new version of circuit svg
* @return Node      Created DOM node
*/
ControllerCircuitDisplay.prototype.createSvg = function() {
    if (!this._cm)
        this._measureCm();

    const bounding = this.node.getBoundingClientRect();
    const width = Math.floor((this.node.parentNode.clientWidth) * 10 / this._cm) - 1;
    const height = Math.floor((this.node.ownerDocument.defaultView.innerHeight - bounding.top) * 10 / this._cm) - 1;
    const offset = this._rearrangeCircuit(width / this.scale - 1, height / this.scale - 1);

    var builder = new ViewCircuitBuilder(
        this.circuit,
        this.node.ownerDocument,
        this.scale,
        width,
        height,
        offset.x,
        offset.y
    );
    builder.addCircuitParts();
    builder.addDots();
    builder.addGrid();
    builder.addSelectors(this);
    return builder.node;
}

/**
* Shedule replacing displayed svg by new svg
* @return undefined
*/
ControllerCircuitDisplay.prototype.updateSvg = function() {
    if (this._replaceTimeout)
        return;
    this._replaceTimeout = window.setTimeout(() => { delete this._replaceTimeout; this._updateSvg(); });
}

/**
* Actually replace displayed svg by new svg
* @return undefined
*/
ControllerCircuitDisplay.prototype._updateSvg = function() {
    var newSvg = this.createSvg();
    this.node.parentNode.replaceChild(newSvg, this.node);
    this.node = newSvg;
}

/**
*/
ControllerCircuitDisplay.prototype.getPlaceClass = function(
    place
) {
    return this._editingMode.procHover.call(this, place, this.circuit);
}

/**
* React on choosing a place
* @return undefined
*/
ControllerCircuitDisplay.prototype.procSelected = function(
    place,
    shiftKey,   // Is shift pressed?
    ctrlKey,    // Is ctrl pressed?
    altKey      // Is alt pressed?
) {
    this._editingMode.procSelected.call(this, place, this.circuit, shiftKey, ctrlKey, altKey);
}
