/**
* A single equation in a system of linear equations
*/
function LinearEquation() {
    this._numbers = [];
}

/** All coefficients and constant */
LinearEquation.prototype._numbers = [];

/** Number of coefficients and constant in equation */
Object.defineProperty(LinearEquation.prototype, 'length', {
    enumerable: true,
    get() { return this._numbers.length; }
});

/**
* Create a new and independent copy of this instance
* @return LinearEquation        Cloned instance
*/
LinearEquation.prototype.clone = function() {
    const clone = new LinearEquation();
    this._numbers.forEach((value, index) => clone.setNumber(index, value));
    return clone;
}

/**
* Get coefficient or constant
* @return Number    Coefficient or constant
*/
LinearEquation.prototype.getNumber = function(
    index       // Index of requested coefficient or constant
) {
    return (index < this._numbers.length) ? this._numbers[index] : 0;
}

/**
* Set coefficient or constant
* @return undefined
*/
LinearEquation.prototype.setNumber = function(
    index,      // Index of coefficient or constant to set
    value       // New value
) {
    while (index > this._numbers.length)
        this._numbers.push(0);
    this._numbers[index] = value;
}

/**
* Multiply linear equation to make leading coefficient 1
* @return Boolean       Did operation succeed?
*/
LinearEquation.prototype.normalize = function(
    index       // Index of coefficient that should be set to 1 (all previouse coefficients must be 0)
) {
    const leading = this.getNumber(index);
    if (!leading)
        return false;
    this._numbers[index] = 1;
    this._numbers.forEach((value, n, arr) => { if (n > index) arr[n] = value / leading; });
}

/**
* Get index of first non zero coefficient
* @return Number    Index
*/
LinearEquation.prototype.getFirstNotZeroIndex = function() {
    for (let n = 0; n < this.length; ++n) {
        if (this._numbers[n])
            return n;
    }
    return Infinity;
}

/**
* Count coefficients beeing 0
* @return Number    Number of coefficients beeing 0
*/
LinearEquation.prototype.countZeros = function() {
    return this._numbers.reduce((prev, coefficient) => (prev + (coefficient ? 0 : 1)), 0);
}

/**
* Subtract multiple of normalized LinearEquation instance equation to obtain a further coefficient 0 (if it is not equal 0 yet)
* @return undefined
*/
LinearEquation.prototype.simplify = function(
    equation        // Normalized instance of LinearEquation
) {
    const index = equation.getFirstNotZeroIndex();
    const factor = this.getNumber(index);
    if (!factor)
        return;
    for (let n = index + 1; n < equation.length; ++n)
        this.setNumber(n, this.getNumber(n) - factor * equation.getNumber(n));
    this.setNumber(index, 0);
}

/**
* Checks, whether or not that single equation can be fullfilled
* @return Boolean       Can equation be fullfilled?
*/
LinearEquation.prototype.isConflict = function() {
    return ((this.getFirstNotZeroIndex() == this.length - 1) && this.getConstant());
}

/**
* Checks, whether or not this equation is 0 = 0
* @return Boolean       Does that equation just say 0 = 0?
*/
LinearEquation.prototype.isTrivial = function() {
    return !this._numbers.reduce((prev, curr) => (prev && !curr), true);
}

/**
* Checks, whether equation is x_index = c for a given single index
* @return Boolean       Is equation x_i = c for a given single index?
*/
LinearEquation.prototype.isSolutionFor = function(
    index       // For what variable?
) {
    if (index >= this.length - 1)
        return false;
    return this._numbers.slice(0, -1).reduce((prev, curr, i) => (prev && ((i == index) == (curr ? true : false))), true);
}

/**
* Gets contstant
* @return Number        Constant
*/
LinearEquation.prototype.getConstant = function() {
    return this.getNumber(this.length - 1);
}

/**
* Create space seperated string of coefficients and constant
* @return String        All values seperated by spaces
*/
LinearEquation.prototype.toString = function() {
    return this._numbers.join(' ');
}
