/**
* A collection of linear equations
*/
function LinearEquationSystem() {
    this._equations = [];
    this._diffs = [];
}

/** All equations */
LinearEquationSystem.prototype._equations = [];

/**
* Add equation
* @return undefined
*/
LinearEquationSystem.prototype.add = function(
    equation        // Equation to add
) {
    this._equations.push(equation);
}

/**
* Checks if the equation with index index2 is more suitable to be solved for the variable col than the equation with index index1
* @return Boolean   Should equation having index index2 be prefered to equation having index index1?
*/
LinearEquationSystem.prototype._cmpEquations = function(
    col,        // Index of coefficient to look at
    index1,     // First equation defined by index in this._equations
    index2      // Second equation defined by index in this._equations
) {
    const eq1 = this._equations[index1];
    const eq2 = this._equations[index2];
    const c1 = eq1.getNumber(col);
    const c2 = eq2.getNumber(col);
    if (!c2)
        return false;
    if (!c1)
        return true;
    const isAbsOne1 = (Math.abs(c1) == 1);
    const isAbsOne2 = (Math.abs(c2) == 1);
    if (isAbsOne1 != isAbsOne2)
        return isAbsOne2;
    const zeros1 = eq1.countZeros();
    const zeros2 = eq2.countZeros();
    if (isAbsOne1 && isAbsOne2) {
        if (zeros1 != zeros2)
            return (zeros1 < zeros2);
        return ((c1 < 0) && (c2 > 0));
    }
    if (zeros1 != zeros2)
        return (zeros1 < zeros2);
    const abweichung1 = Math.abs(Math.abs(c1) - 1);
    const abweichung2 = Math.abs(Math.abs(c2) - 1);
    return (abweichung1 < abweichung2);
}

/**
* Rearrange equations with index greater or equal row to have equation with greatest absolute value (or even better: +1 or -1) of col-th coefficient row-th equation
* @return Number        Greatest col-th coefficient (absolute value)
*/
LinearEquationSystem.prototype._exchange = function(
    row,        // Index of first equation to consider
    col         // Index of coefficient to look at
) {
    if (row >= this._equations.length)
        return 0;
    let row_best = row;
    for (let r = row + 1; r < this._equations.length; ++r) {
        if (this._cmpEquations(col, row_best, r))
        row_best = r;
    }
    if (row_best != row) {
        const tmp = this._equations[row_best];
        this._equations[row_best] = this._equations[row];
        this._equations[row] = tmp;
    }
    return this._equations[row].getNumber(col);
}

/**
* Calculate maximum number of coefficients plus constant per equation
* @return Number    Maximum number
*/
LinearEquationSystem.prototype.countCols = function() {
    return this._equations.map(equation => equation.length - 1).reduce((a, b) => Math.max(a, b), 0);
}

/**
* Remove equations that are just a multiple of an previouse equation
* @return undefined
*/
LinearEquationSystem.prototype.removeDuplicates = function() {
    const arrFingerprints = [];
    this._equations = this._equations.filter(equation => {
        const clone = equation.clone();
        clone.normalize();
        const fingerprint = clone.toString();
        const noDuplicate = (arrFingerprints.indexOf(fingerprint) == -1);
        if (noDuplicate)
            arrFingerprints.push(fingerprint);
        return true;
    });
}

/**
* Tries to solve linear equation system
* @return undefined
*/
LinearEquationSystem.prototype.solve = function() {
    // Ensure equal length of equations
    const maxlen = this.countCols();
    this._equations.filter(equation => (equation.length <= maxlen)).forEach(equation => equation.setNumber(maxlen, 0));
    // Remove equations that are just a multiple of an previouse equation
    this.removeDuplicates();
    // Ensure triangle shape
    let col = 0;
    let row = 0;
    while ((row < this._equations.length) && (col < maxlen)) {
        if (this._exchange(row, col)) {
            this._equations[row].normalize(col);
            this._equations.forEach((equation, index) => { if (index > row) equation.simplify(this._equations[row]); });
            ++row;
        }
        ++col;
    }
    // Solve equations
    this._equations.forEach((equation, index, arr) => {
        for (let i = index + 1; i < arr.length; ++i)
            equation.simplify(arr[i]);
    });
}

/**
* Checkes (after solving!) whether or not there is a conflict in the linear equation system
* @return Boolean       Is there a conflict?
*/
LinearEquationSystem.prototype.isConflict = function() {
    return this._equations.reduce((prev, curr) => (prev || curr.isConflict()), false);
}

/**
* Reads solution for single variable from solved linear equation system
* @return Number        Solution (or undefined)
*/
LinearEquationSystem.prototype.getSolution = function(
    index       // Number of variable
) {
    for (let n = 0; n < this._equations.length; ++n) {
        if (this._equations[n].isSolutionFor(index))
            return this._equations[n].getConstant();
    }
    return undefined;
}

// Cached differences
LinearEquationSystem.prototype._diffs = [];

/**
* Calculate difference between solution for minuend-th and solution for subtrahend-th variable from solved linear equation system
* @return Number        Difference (or undefined if unable to calculate)
*/
LinearEquationSystem.prototype.calcDifference = function(
    minuend,        // Index/number of variable to be subtracted
    subtrahend      // Index/number of variable to subtract
) {
    if ((this._diffs[minuend] instanceof Array) && (subtrahend < this._diffs[minuend].length) && !Number.isNaN(this._diffs[minuend][subtrahend]))
        return this._diffs[minuend][subtrahend];
    while (this._diffs.length < minuend)
        this._diffs.push(NaN);
    if (!(this._diffs[minuend] instanceof Array))
        this._diffs[minuend] = [];
    while (this._diffs[minuend].length < subtrahend)
        this._diffs[minuend].push(NaN);
    const result = this._calcDifference(minuend, subtrahend);
    this._diffs[minuend][subtrahend] = result;
    return result;
}

/**
* Calculate difference between solution for minuend-th and solution for subtrahend-th variable from solved linear equation system by solving another equation system
* @return Number        Difference (or undefined if unable to calculate)
*/
LinearEquationSystem.prototype._calcDifference = function(
    minuend,        // Index/number of variable to be subtracted
    subtrahend      // Index/number of variable to subtract
) {
    if (minuend == subtrahend)
        return 0;
    const minuendValue = this.getSolution(minuend);
    const subtrahendValue = this.getSolution(subtrahend);
    if (Number.isFinite(minuendValue) && Number.isFinite(subtrahendValue))
        return minuendValue - subtrahendValue;
    if (Number.isFinite(minuendValue) != Number.isFinite(subtrahendValue))
        return undefined;
    const subsystem = new LinearEquationSystem();
    this._equations.map(equation => subsystem.add(equation.clone()));
    const ground = new LinearEquation();
    ground.setNumber(subtrahend, 1);
    subsystem.add(ground);
    subsystem.solve();
    return subsystem.getSolution(minuend);
}

/**
* Converts linear equation system to matrix representation
* @return String        Matrix representation
*/
LinearEquationSystem.prototype.toString = function() {
    const data = this._equations.map(equation => equation._numbers.map(v => ('' + v)));
    const itemCount = data.reduce((prev, curr) => Math.max(prev, curr.length), 0);
    if (itemCount < 2)
        return '';
    data.forEach(line => { while (line.length < itemCount) line.push('0')});
    data.forEach(line => { const last = line.pop(); line.push('|'); line.push(last); });
    const len = data.map(line => line.map(s => s.length)).reduce((prev, curr) => prev.map((max, index) => Math.max(max, curr[index])));
    return data.map(line => line.map((s, index) => (' '.repeat(len[index] - s.length) + s)).join(' ')).join('\n');
}
