/**
* Data of a changeable and simulatable circuit diagram
*/
function ModelCircuit() {
    this._parts = [];
    this.eventManager = new MultiListenerManager(this);
}

/** All components of this circuit diagram */
ModelCircuit.prototype._parts = [];

/** Object managing ModelCircuit's events */
ModelCircuit.prototype.eventManager = null;

/** Is simulation of circuit active? */
ModelCircuit.prototype._simulationSheduleManager = null;
Object.defineProperty(ModelCircuit.prototype, 'simulating', {
    enumerable: true,
    get() { return this._simulationSheduleManager ? true : false; },
    set(v) {
        const realValue = (v ? true : false);
        if (realValue == this.simulating)
            return;
        if (realValue) {
            this._simulationSheduleManager = new SimulationSheduleManager();
            this.eventManager.fireEvent('simulationStart');
            this._simulate();
        } else {
            this._simulationSheduleManager.stop();
            delete this._simulationSheduleManager;
            this._parts.forEach(part => part.setSimulationStatus(undefined, undefined, this._simulationSheduleManager));
            this.eventManager.fireEvent('simulationEnd');
        }
    }
});

/**
* Delete all components located at the specified grid line
* @return Array     All deleted ModelCircuitPart instances
*/
ModelCircuit.prototype.clearGridLine = function(
    x,              // Lowest horizontal coordinate of grid point corresponding to grid line
    y,              // Lowest vertical coordinate of grid point corresponding to grid line
    isHorizontal    // Is it a horizontal grid line?
) {
    var arrDelete = this._parts.filter(function(part) { return part.isAtGridPos(x, y, isHorizontal); });
    this._parts = this._parts.filter(function(part) { return !part.isAtGridPos(x, y, isHorizontal); });
    arrDelete.map(function(part) { part.component.bubbleEvent('deleted'); });
    return arrDelete;
}

/**
* Delete all components
* @return Array     All deleted ModelCircuitPart instances
*/
ModelCircuit.prototype.clear = function() {
    var arrDelete = this._parts;
    this._parts = [];
    this.eventManager.fireEvent('cleared', arrDelete);
    return arrDelete;
}

/**
* Insert a component at the specified grid line
* @return ModelCircuitPart      Created ModelCircuitPart instance
*/
ModelCircuit.prototype.addComponent = function(
    component,      // ModelComponent instance that should be part of ModelCircuit
    x,              // Lowest horizontal coordinate of grid point corresponding to a component's connector
    y,              // Lowest vertical coordinate of grid point corresponding to a component's connector
    isHorizontal,   // Do primary and secondary connector share the same vertical position?
    isFlipped       // Does the component's secondary connector have a lower horizontal or vertical coordinate of its grid point?
) {
    if (component.circuit != this)
        return null;
    this.clearGridLine(x, y, isHorizontal);
    var part = new ModelCircuitPart(component, x, y, isHorizontal, isFlipped);
    this._parts.push(part);
    component.bubbleEvent('inserted');
    return part;
}

/**
* Calculate area of grid used by all parts
* @return Object    { left: "Lowest horizontal coordinate", right: "Highest horizontal coordinate", top: "Lowest vertical coordinate", bottom: "Highest vertical coordinate" }
*/
ModelCircuit.prototype.getUsedGridArea = function() {
    if (!this._parts.length)
        return { left: 0, right: 0, top: 0, bottom: 0 };
    return this._parts.map(
        function(part) { return part.getUsedGridArea(); }
    ).reduce(
        function(rect1, rect2) {
            return { left: Math.min(rect1.left, rect2.left), right: Math.max(rect1.right, rect2.right), top: Math.min(rect1.top, rect2.top), bottom: Math.max(rect1.bottom, rect2.bottom) };
        }
    );
}

/**
* Notify event listeners that a change has occured and simulate (if simulating is active)
* @return undefined
*/
ModelCircuit.prototype.bubbleEvent = function(
    what,       // Name of occured event
    component   // Component involved
) {
    this.eventManager.fireEvent(what, component);
    if (this.simulating)
        this._simulate();
}

/**
* Call function fkt on every circuit part
* @return Array     Results of mapping
*/
ModelCircuit.prototype.map = function(
    fkt     // Function to call
) {
    return this._parts.map(fkt);
}

/**
* Find all circuit parts beeing connected to a grid point
* @return Array     Array of all found ModelCircuitPart instances
*/
ModelCircuit.prototype.getPartsConnectedTo = function(
    x,              // Horizontal coordinate of grid point corresponding to a component's connector
    y               // Vertical coordinate of grid point corresponding to a component's connector
) {
    return this._parts.filter(function(part) {
        return part.hasConnectorAt(x, y);
    });
}

/**
* Find circuit part at specified place
* @return ModelCircuitPart      Circuit part at specified place (or null)
*/
ModelCircuit.prototype.getElementAt = function(
    x,              // Lowest horizontal coordinate of grid point corresponding to a component's connector
    y,              // Lowest vertical coordinate of grid point corresponding to a component's connector
    isHorizontal    // Do primary and secondary connector share the same vertical position?
) {
    return this._parts.filter(function(part) {
        return part.isAtGridPos(x, y, isHorizontal);
    }).reduce(function(prev, next) {
        return prev || next;
    }, null);
}

/**
* Create a string representing circuit
* @return String    Serialized circuit
*/
ModelCircuit.prototype.serialize = function() {
    var area = this.getUsedGridArea();
    var data = [];
    var x, y, part;
    for (y = area.top; y <= area.bottom; ++y) {
        for (x = area.left; x < area.right; ++x) {
            part = this.getElementAt(x, y, true);
            data.push(part && part.serialize());
        }
        if (y != area.bottom) {
            for (x = area.left; x <= area.right; ++x) {
                part = this.getElementAt(x, y, false);
                data.push(part && part.serialize());
            }
        }
    }
    if (!data.length)
        return '0';
    return (area.right - area.left) + ' ' + data.map(function(v) {
        return ((typeof v == 'string') || (v instanceof String)) ? v : '';
    }).reduce(
        function(arr, s) {
            if (arr.length && (arr[arr.length - 1].s == s))
                ++arr[arr.length - 1].n;
            else
                arr.push({ n: 1, s: s });
            return arr;
        }, []
    ).map(
        function(o) { return (o.n > 1) ? (o.n - 2 + o.s) : o.s; }
    ).join(' ');
}

/**
* Add single component from serialized string
* @return undefined
*/
ModelCircuit.prototype.addSerialized = function(
    x,              // Lowest horizontal coordinate of grid point corresponding to a component's connector
    y,              // Lowest vertical coordinate of grid point corresponding to a component's connector
    isHorizontal,   // Do primary and secondary connector share the same vertical position?
    s               // Serialized string describing component
) {
    var firstLetter = s.substr(0, 1);
    var serializingLetter = firstLetter.toLowerCase();
    var componentClass = ModelComponent.serializingLetters[serializingLetter];
    if (!componentClass)
        return;
    var component = componentClass.fromSerialized(this, serializingLetter + s.substr(1));
    if (!component)
        return;
    this.addComponent(component, x, y, isHorizontal, firstLetter != serializingLetter);
}

/**
* Load circuit from serialized string
* @return Boolean       Loading succeded?
*/
ModelCircuit.prototype.unserialize = function(
    s           // Serialized circuit as a string
) {
    var muted = this.eventManager.muted;
    this.eventManager.muted = true;
    this.clear();
    var data = s.split(' ');
    var width = +data.shift();
    if (!Number.isFinite(width) || (width < 0))
        return false;
    var regexpDigits = new RegExp('^[0-9]+');
    var parts = data.map(
        function(s) {
            var match = s.match(regexpDigits);
            return match ? { n: +match[0] + 2, s: s.substr(match[0].length) } : { n: 1, s: s };
        }
    );
    var x = 0;
    var y = 0;
    var isHorizontal = (width > 0);
    while (parts.length) {
        this.addSerialized(x, y, isHorizontal, parts[0].s);
        if (--parts[0].n < 1)
            parts.shift();
        if (++x > (isHorizontal ? (width - 1) : width)) {
            x = 0;
            if (!width || (isHorizontal = !isHorizontal))
                ++y;
        }
    }
    this.eventManager.muted = muted;
    this.eventManager.fireEvent('restored');
    return true;
}

/**
* Generate part arrays of this._parts containing only connected components
* @return Array     Total disjunctive arrays of connected components
*/
ModelCircuit.prototype.getArraysOfConnectedParts = function() {
    var size = this.getUsedGridArea();
    var width = Math.max(1, 1 + size.right - size.left);
    function nameConnection(o) {
        return o.x + o.y * width;
    }
    var result = this.map(function(part) {
        return { parts: [ part ], connections: part.getAllConnectors().map(nameConnection) };
    });
    result.forEach(function(obj, index, array) {
        for (var n = 0; n < index; ++n) {
            if (array[n] && obj.connections.reduce(function(prev, curr) {
                return prev || (array[n].connections.indexOf(curr) != -1)
            }, false)) {
                obj.parts = obj.parts.concat(array[n].parts);
                array[n].connections.forEach(function(connection) {
                    if (obj.connections.indexOf(connection) == -1)
                        obj.connections.push(connection);
                });
                array[n] = false;
            }
        }
    })
    return result.filter(function(v) { return (v !== false); }).map(function(o) { return o.parts; });
}

/**
* Create SimulationCircuitArea instances for simulating the hole circuit
* @return Array     Array of created SimulationCircuitArea instances
*/
ModelCircuit.prototype.createSimulationCircuitArea = function() {
    const width = GridPointHasher.calcWidth(this);
    let arrBatteries = this._parts.filter(part => part.getConstraintVoltage());
    let result = [];
    while (arrBatteries.length) {
        const sca = new SimulationCircuitArea(width);
        sca.build(arrBatteries.pop(), this);
        result.push(sca);
        arrBatteries = arrBatteries.filter(part => (sca.parts.indexOf(part) == -1));
    }
    return result;
}

/**
* Simulate circuit area and apply calculated currents and short circuits
* @return undefined
*/
ModelCircuit.prototype._simulate = function() {
    const areas = this.createSimulationCircuitArea();
    areas.forEach(area => area.simulate(this._simulationSheduleManager));
    const simulatedParts = areas.map(area => area.parts).reduce((prev, curr) => prev.concat(curr), []);
    this._parts.filter(part => (simulatedParts.indexOf(part) == -1)).forEach(part => part.setSimulationStatus(0, 0, this._simulationSheduleManager));
    this.eventManager.fireEvent('simulationStep');
    this._simulationSheduleManager.start();
}

/**
* Get a CircuitMesh instance for each mesh containing what
* @return Array      Array of CircuitMesh instances
*/
ModelCircuit.prototype.getMeshes = function(
    what,           // Filter for start component of mesh (specific ModelCircuitPart or ModelComponent instance, ModelComponent class represented by class or letter (see ModelComponent.serializingLetters), object with required pairs of key and value or specified by individual filter function acceptOnly)
    keepIncomplete  // Optional: Keep discontinued meshes
) {
    const builder = new CircuitMeshBuilder(this);
    return this._parts.filter(ModelCircuitPart.createFilterFunction(what)).map(part => builder.buildMeshes(part, keepIncomplete)).reduce((prev, curr) => prev.concat(curr), []);
}
