/**
* Class for entry of ModelCircuit: Position and information on a component
*/
function ModelCircuitPart(
    component,      // ModelComponent instance that should be part of ModelCircuit
    x,              // Lowest horizontal coordinate of grid point corresponding to a component's connector
    y,              // Lowest vertical coordinate of grid point corresponding to a component's connector
    isHorizontal,   // Do primary and secondary connector share the same vertical position?
    isFlipped       // Does the component's secondary connector have a lower horizontal or vertical coordinate of its grid point?
) {
    Object.defineProperty(this, 'component', { enumerable: true, writeable: false, value: component });
    this.setPosition(x, y, isHorizontal, isFlipped);
}

// Assigned ModelComponent instance
ModelCircuitPart.prototype.component = null;

// Lowest horizontal coordinate of grid point corresponding to a component's connector
ModelCircuitPart.prototype._x = 0;
Object.defineProperty(ModelCircuitPart.prototype, 'x', { enumerable: true,
    get() { return this._x; },
    set(v) { this.setPosition(v, this._y, this._isHorizontal, this._isFlipped); }
});

// Lowest vertical coordinate of grid point corresponding to a component's connector
ModelCircuitPart.prototype._y = 0;
Object.defineProperty(ModelCircuitPart.prototype, 'y', { enumerable: true,
    get() { return this._y; },
    set(v) { this.setPosition(this._x, v, this._isHorizontal, this._isFlipped); }
});

// Do primary and secondary connector share the same vertical position?
ModelCircuitPart.prototype._isHorizontal = false;
Object.defineProperty(ModelCircuitPart.prototype, 'isHorizontal', { enumerable: true,
    get() { return this._isHorizontal; },
    set(v) { this.setPosition(this._x, this._y, v, this._isFlipped); }
});

// Does the component's secondary connector have a lower horizontal or vertical coordinate of its grid point?
ModelCircuitPart.prototype._isFlipped = false;
Object.defineProperty(ModelCircuitPart.prototype, 'isFlipped', { enumerable: true,
    get() { return this._isFlipped; },
    set(v) { this.setPosition(this._x, this._y, this._isHorizontal, v); }
});

/**
* Returns whether this component is symmetrical
* @return Boolean   Does the orientation of this component *not* make any difference?
*/
ModelCircuitPart.prototype.isSymmetric = function() {
    return this.component.isSymmetric();
}

/**
* Returns whether the user can interact with this component (e.g. turning a switch on and off, change a sensor's value or turn a potentiometer)
* @return Boolean   Can user interact with this component?
*/
ModelCircuitPart.prototype.isInteractive = function() {
    return this.component.isInteractive();
}

/**
* Interact with this component (e.g. turn switch on or off, change sensor's value, …)
* @return undefined
*/
ModelCircuitPart.prototype.interactWith = function() {
    this.component.interactWith();
}

/**
* Returns whether the state or subtype of the component can be set
* @return Boolean   Can user change preferences of this component?
*/
ModelCircuitPart.prototype.isConfigureable = function() {
    return this.component.isConfigureable();
}

/**
* Start configuration of this component
* @return undefined
*/
ModelCircuitPart.prototype.startConfiguration = function() {
    this.component.startConfiguration();
}

/**
* Notify component on position change
* @return undefined
*/
ModelCircuitPart.prototype._notify = function() {
    if (this.component instanceof ModelComponent)
        this.component.bubbleEvent('positionChange');
}

/**
* Change position of assigned component and notify
* @return undefined
*/
ModelCircuitPart.prototype.setPosition = function(
    x,              // Lowest horizontal coordinate of grid point corresponding to a component's connector
    y,              // Lowest vertical coordinate of grid point corresponding to a component's connector
    isHorizontal,   // Do primary and secondary connector share the same vertical position?
    isFlipped       // Does the component's secondary connector have a lower horizontal or vertical coordinate of its grid point?
) {
    var self = this;
    if ([
        { key: '_x', value: x, canchange: true },
        { key: '_y', value: y, canchange: true },
        { key: '_isHorizontal', value: (isHorizontal ? true : false), canchange: true },
        { key: '_isFlipped', value: (isFlipped ? true : false), canchange: !this.isSymmetric() }
    ].filter(
        function(item) { return (self[item.key] != item.value) && item.canchange; }
    ).map(
        function(item) { self[item.key] = item.value; }
    ).length)
        this._notify();
}

/**
* Checks whether the assigned component is located at the specified grid line
* @return boolean   Is assigned component located at specified grid line?
*/
ModelCircuitPart.prototype.isAtGridPos = function(
    x,              // Lowest horizontal coordinate of grid point corresponding to grid line
    y,              // Lowest vertical coordinate of grid point corresponding to grid line
    isHorizontal    // Is it a horizontal grid line?
) {
    if (this._x != x)
        return false;
    if (this._y != y)
        return false;
    if (this._isHorizontal != (isHorizontal ? true : false))
        return false;
    return true;
}

/**
* Checks whether there is a connector of the assigned component at the specified grid point
* @return boolean   Is assigned component connected to the specified grid point?
*/
ModelCircuitPart.prototype.hasConnectorAt = function(
    x,          // Horizontal coordinate of grid point
    y           // Vertical coordinate of grid point
) {
    if ((this._x == x) && (this._y == y))
        return true;
    if (this._isHorizontal && (this._x + 1 == x) && (this._y == y))
        return true;
    if (!this._isHorizontal && (this._x == x) && (this._y + 1 == y))
        return true;
    return false;
}

/**
* Return array of all grid points this component ist connected to
* @return Array     [ {x, y}, … ]
*/
ModelCircuitPart.prototype.getAllConnectors = function() {
    return [ { x: this.x, y: this.y}, { x: this.x + (this._isHorizontal ? 1 : 0), y: this.y + (this._isHorizontal ? 0 : 1) } ];
}

/**
* Calculate area of grid used by this part
* @return Object    { left: "Lowest horizontal coordinate", right: "Highest horizontal coordinate", top: "Lowest vertical coordinate", bottom: "Highest vertical coordinate" }
*/
ModelCircuitPart.prototype.getUsedGridArea = function() {
    return { left: this._x, top: this._y, right: this._x + (this._isHorizontal ? 1 : 0), bottom: this._y + (this._isHorizontal ? 0 : 1) };
}

/**
* Create string representing circuit part in serialized string
* @return String        String representing circuit part (starting with corresponding letter registered in ModelComponent.serializingLetters) or false if component cannot be serialized
*/
ModelCircuitPart.prototype.serialize = function() {
    var s = this.component.serialize();
    if ((s === false) || !this._isFlipped)
        return s;
    return s.substr(0, 1).toUpperCase() + s.substr(1);
}

/**
* Get present constant resistance of this component (if any)
* @return Number    Resistance in ohms (can be infinity)
*/
ModelCircuitPart.prototype.getConstraintResistance = function() {
    return this.component.getConstraintResistance();
}

/**
* Get present constant voltage generated by this component (if any)
* @return Number    Voltage in volts (0 if this component does not generate a voltage); positive number means secondary connector has a higher potential than primary connector
*/
ModelCircuitPart.prototype.getConstraintVoltage = function() {
    return this.component.getConstraintVoltage() * (this.isFlipped ? -1 : 1);
}

/**
* Returns whether both connections are not electrical exchangeable
* @return Boolean   Are there different kinds of connectors?
*/
ModelCircuitPart.prototype.isDifferentConnectors = function() {
    return this.component.isDifferentConnectors();
}

/**
* Set current simulation status (e. g. current)
* @return undefined
*/
ModelCircuitPart.prototype.setSimulationStatus = function(
    current,        // Current through component (undefined → no simulation; Infinity → short circuit)
    voltage,        // Voltage between connections of this component; positive number means secondary connector has a higher potential than primary connector
    sheduleManager  // Responsible SimulationSheduleManager instance
) {
    if (this.isFlipped && Number.isFinite(current) && current)
        this.component.setSimulationStatus(-current, -voltage, sheduleManager);
    else
        this.component.setSimulationStatus(current, voltage, sheduleManager);
}

/**
* Creates a function that can be used to filter an array of ModelCircuitPart instances accepting only what is described by acceptOnly
* @return Function      Function suitable for Array.prototype.filter
*/
ModelCircuitPart.createFilterFunction = function(
    acceptOnly      // What to accept (specific ModelCircuitPart or ModelComponent instance, ModelComponent class represented by class or letter (see ModelComponent.serializingLetters), object with required pairs of key and value or specified by individual filter function acceptOnly)
) {
    if (acceptOnly instanceof ModelCircuitPart)
        return ((part) => (part == acceptOnly));
    const fktFilterComponent = ModelComponent.createFilterFunction(acceptOnly);
    return (function(part) {
        return fktFilterComponent.call(this, part.component);
    });
}
