/**
* Abstract class for all types of components within a circuit diagram
*/
function ModelComponent(
    circuit     // ModelCircuit instance this instance should belong to
) {
    Object.defineProperty(this, 'circuit', { enumerable: true, writeable: false, value: circuit });
}

/** ModelCircuit instance of which this component is part of */
ModelComponent.prototype.circuit = null;

/**
* Notify ModelCircuit instance that there was a change to this component
* @return undefined
*/
ModelComponent.prototype.bubbleEvent = function(
    what        // Name of occured event
) {
    if (this.circuit instanceof ModelCircuit)
        this.circuit.bubbleEvent(what, this);
}

/**
* Get present constant resistance of this component (if any)
* @return Number    Resistance in ohms (can be infinity)
*/
ModelComponent.prototype.getConstraintResistance = function() {
    return Infinity;
}

/**
* Get present constant voltage generated by this component (if any)
* @return Number    Voltage in volts (0 if this component does not generate a voltage); positive number means secondary connector has a higher potential than primary connector
*/
ModelComponent.prototype.getConstraintVoltage = function() {
    return 0;
}

/**
* Returns whether this component is symmetrical
* @return Boolean   Does the orientation of this component *not* make any difference?
*/
ModelComponent.prototype.isSymmetric = function() {
    return true;
}

/**
* Returns whether both connections are not electrical exchangeable
* @return Boolean   Are there different kinds of connectors?
*/
ModelComponent.prototype.isDifferentConnectors = function() {
    return false;
}

/**
* Returns whether the user can interact with this component during simulation (e.g. turning a switch on and off, change a sensor's value or turn a potentiometer)
* @return Boolean   Can user interact with this component?
*/
ModelComponent.prototype.isInteractive = function() {
    return false;
}

/**
* Interact with this component during simulation (e.g. turn switch on or off, change sensor's value, …)
* @return undefined
*/
ModelComponent.prototype.interactWith = function() {
}

/**
* Returns whether the state or subtype of the component can be set
* @return Boolean   Can user change preferences of this component?
*/
ModelComponent.prototype.isConfigureable = function() {
    return false;
}

/**
* Start configuration of this component
* @return undefined
*/
ModelComponent.prototype.startConfiguration = function() {
}

/**
* Create string representing component in serialized string
* @return String        String representing component (starting with corresponding letter registered in ModelComponent.serializingLetters) or false if component cannot be serialized
*/
ModelComponent.prototype.serialize = function() {
    return false;
}

/**
* Which letter in serialized string correspondes to which child class of ModelComponent? Only lowercase is allowed, flippable components are represented by corresponding uppercase letter when flipped
*/
ModelComponent.serializingLetters = {};

/**
* Creates a function that can be used to filter an array of ModelComponent instances accepting only what is described by acceptOnly
* @return Function      Function suitable for Array.prototype.filter
*/
ModelComponent.createFilterFunction = function(
    acceptOnly      // What to accept (specific ModelComponent instance, ModelComponent class represented by class or letter (see ModelComponent.serializingLetters), object with required pairs of key and value or specified by individual filter function acceptOnly)
) {
    switch (true) {
        case (acceptOnly instanceof ModelComponent):
            return ((component) => (component == acceptOnly));
        case ((acceptOnly instanceof Object) && (acceptOnly.prototype instanceof ModelComponent)):
            return ((component) => (component instanceof acceptOnly));
        case ((typeof acceptOnly == 'string') || (acceptOnly instanceof String)):
            const componentClass = ModelComponent.serializingLetters[acceptOnly.charAt(0).toLowerCase()];
            return componentClass ? ((component) => (component instanceof acceptOnly)) : (() => false);
        case (typeof acceptOnly == 'function'):
            return acceptOnly;
        case (acceptOnly instanceof Object):
            return ((component) => {
                for (let key in acceptOnly) {
                    if (((typeof component[key] == 'function') ? component[key]() : component[key]) != acceptOnly[key])
                        return false;
                }
                return true;
            });
        default:
            return (() => false);
    }
}

/**
* Set current simulation status (e. g. current)
* @return undefined
*/
ModelComponent.prototype.setSimulationStatus = function(
    current,        // Current through component (undefined → no simulation; Infinity → short circuit)
    voltage,        // Voltage between connections of this component; positive number means secondary connector has a higher potential than primary connector
    sheduleManager  // Responsible SimulationSheduleManager instance
) {
}
