/**
* A fuse being part of a circuit
*/
function ModelFuse(
    circuit     // ModelCircuit instance this instance should belong to
) {
    ModelFuse.parent.call(this, circuit);
}
ModelFuse.parent = ModelComponent;
ModelFuse.prototype = Object.create(ModelFuse.parent.prototype);
ModelFuse.prototype.constructor = ModelFuse;

/** Is fuse blown within the simulation? */
ModelFuse.prototype._isBlown = false;
Object.defineProperty(ModelFuse.prototype, 'isBlown', { enumerable: true,
    get() { return this._isBlown; },
    set(v) {
        const realValue = v ? true : false;
        if (realValue == this._isBlown)
            return;
        this._isBlown = realValue;
        if (realValue)
            this.bubbleEvent('simulationFuseReset');
        else
            this.bubbleEvent('simulationFuseBlown');
    }
});

/** Time delay in milli seconds before fuse blows */
ModelFuse.prototype.delay = 100;

/** Id of timeout for blowing */
ModelFuse.prototype._timeoutId = undefined;

/**
* Get present constant resistance of this component (if any)
* @return Number    Resistance in ohms (can be infinity)
*/
ModelFuse.prototype.getConstraintResistance = function() {
    return this.isBlown ? Infinity : 0;
}

/**
* Create string representing component in serialized string
* @return String        String representing component (starting with corresponding letter registered in ModelComponent.serializingLetters) or false if component cannot be serialized
*/
ModelFuse.prototype.serialize = function() {
    return 'f';
}

/**
* Which letter in serialized string correspondes to which child class of ModelComponent? Only lowercase is allowed, flippable components are represented by corresponding uppercase letter when flipped
*/
ModelComponent.serializingLetters['f'] = ModelFuse;

/**
* Create ModelFuse instance from serialized string s
* @return ModelFuse    Instance described by serialized string
*/
ModelFuse.fromSerialized = function(
    circuit,    // ModelCircuit instance this instance should belong to
    s           // Serialized string describing component
) {
    return new ModelFuse(circuit);
}

/**
* Set current simulation status (e. g. current)
* @return undefined
*/
ModelFuse.prototype.setSimulationStatus = function(
    current,        // Current through component (undefined → no simulation; Infinity → short circuit)
    voltage,        // Voltage between connections of this component; positive number means secondary connector has a higher potential than primary connector
    sheduleManager  // Responsible SimulationSheduleManager instance
) {
    switch (true) {
        case (current === undefined):
            this._isBlown = false;
            if (this._timeoutId !== undefined) {
                sheduleManager.remove(this._timeoutId);
                delete this._timeoutId;
            }
            break;
        case (current === Infinity):
            if ((this._timeoutId === undefined) && !this.isBlown)
                this._timeoutId = sheduleManager.add(() => { delete this._timeoutId; this.isBlown = true; }, this.delay);
            break;
        default:
            if (this._timeoutId !== undefined) {
                sheduleManager.remove(this._timeoutId);
                delete this._timeoutId;
            }
            break;
    }
    this.shortCircuited = (Math.abs(current) == Infinity);
}

/**
* Returns whether the user can interact with this component during simulation (e.g. turning a switch on and off, change a sensor's value or turn a potentiometer)
* @return Boolean   Can user interact with this component?
*/
ModelFuse.prototype.isInteractive = function() {
    return this.isBlown;
}

/**
* Interact with this component during simulation (e.g. turn switch on or off, change sensor's value, …)
* @return undefined
*/
ModelFuse.prototype.interactWith = function() {
    this.isBlown = false;
}
