/**
* A meter for voltage or current intensity beeing part of the circuit
*/
function ModelMeter(
    circuit,        // ModelCircuit instance this instance should belong to
    measureCurrent  // Optional: Measure current intensity instead of voltage?
) {
    ModelMeter.parent.call(this, circuit);
    if (measureCurrent)
        this._measureCurrent = true;
}
ModelMeter.parent = ModelComponent;
ModelMeter.prototype = Object.create(ModelMeter.parent.prototype);
ModelMeter.prototype.constructor = ModelMeter;

/** Measure current intensity instead of voltage? */
ModelMeter.prototype._measureCurrent = false;
Object.defineProperty(ModelMeter.prototype, 'measureCurrent', {
    enumerable: true,
    get() { return this._measureCurrent; },
    set(v) {
        var value = (v ? true : false);
        if (value == this._measureCurrent)
            return;
        this._measureCurrent = value;
        this.bubbleEvent('switchMeasurementMethod');
    }
});

/**
* Returns whether the state or subtype of the component can be set
* @return Boolean   Can user change preferences of this component?
*/
ModelMeter.prototype.isConfigureable = function() {
    return true;
}

/**
* Start configuration of this component
* @return undefined
*/
ModelMeter.prototype.startConfiguration = function() {
    this.measureCurrent = !this.measureCurrent;
}

/**
* Get present constant resistance of this component (if any)
* @return Number    Resistance in ohms (can be infinity)
*/
ModelMeter.prototype.getConstraintResistance = function() {
    return this._measureCurrent ? 0 : Infinity;
}

/** Current or voltage measured (calculated by simulation) */
ModelMeter.prototype._measured = undefined;
Object.defineProperty(ModelMeter.prototype, 'measured', {
    enumerable: true,
    get() { return this._measured; },
    set(v) {
        if (v == this._measured)
            return;
        this._measured = v;
//        this.bubbleEvent('measuerementChanged');
    }
});

/**
* Set current simulation status (e. g. current)
* @return undefined
*/
ModelMeter.prototype.setSimulationStatus = function(
    current,        // Current through component (undefined → no simulation; Infinity → short circuit)
    voltage         // Voltage between connections of this component; positive number means secondary connector has a higher potential than primary connector
) {
    if (current == undefined)
        this.measured = undefined;
    else
        this.measured = this._measureCurrent ? current : voltage;
}

/** Types of meter an corresponding letter in serialized strings */
ModelMeter.TYPE_CURRENT = '?';
ModelMeter.TYPE_VOLTAGE = ']';

/**
* Create string representing component in serialized string
* @return String        String representing component (starting with corresponding letter registered in ModelComponent.serializingLetters) or false if component cannot be serialized
*/
ModelMeter.prototype.serialize = function() {
    return this._measureCurrent ? ModelMeter.TYPE_CURRENT : ModelMeter.TYPE_VOLTAGE;
}

/**
* Which letter in serialized string correspondes to which child class of ModelComponent? Only lowercase is allowed, flippable components are represented by corresponding uppercase letter when flipped
*/
ModelComponent.serializingLetters[ModelMeter.TYPE_CURRENT] = ModelMeter;
ModelComponent.serializingLetters[ModelMeter.TYPE_VOLTAGE] = ModelMeter;

/**
* Create ModelMeter instance from serialized string s
* @return ModelMeter    Instance described by serialized string
*/
ModelMeter.fromSerialized = function(
    circuit,    // ModelCircuit instance this instance should belong to
    s           // Serialized string describing component
) {
    return new ModelMeter(circuit, s == ModelMeter.TYPE_CURRENT);
}
