/**
* A horizontal or vertical lightbulb being part of a circuit
*/
function ModelSwitch(
    circuit     // ModelCircuit instance this instance should belong to
) {
    ModelSwitch.parent.call(this, circuit);
}
ModelSwitch.parent = ModelComponent;
ModelSwitch.prototype = Object.create(ModelSwitch.parent.prototype);
ModelSwitch.prototype.constructor = ModelSwitch;

/** Is switch turned on */
ModelSwitch.prototype._on = false;
ModelSwitch.prototype._simulationToggled = false;
Object.defineProperty(ModelSwitch.prototype, 'on', {
    enumerable: true,
    get() { return this._simulationToggled ? (!this._on) : this._on; },
    set(v) {
        var value = (v ? true : false);
        if (value == this._on)
            return;
        this._on = value;
        this.bubbleEvent('switchOperated');
    }
});

/**
* Returns whether the user can interact with this component (e.g. turning a switch on and off, change a sensor's value or turn a potentiometer)
* @return Boolean   Can user interact with this component?
*/
ModelSwitch.prototype.isInteractive = function() {
    return true;
}

/**
* Interact with this component (e.g. turn switch on or off, change sensor's value, …)
* @return undefined
*/
ModelSwitch.prototype.interactWith = function() {
    this._simulationToggled = !this._simulationToggled;
    this.bubbleEvent('simulationSwitchOperated');
}

/**
* Returns whether the state or subtype of the component can be set
* @return Boolean   Can user change preferences of this component?
*/
ModelSwitch.prototype.isConfigureable = function() {
    return true;
}

/**
* Start configuration of this component
* @return undefined
*/
ModelSwitch.prototype.startConfiguration = function() {
    this.on = !this.on;
}

/**
* Get present constant resistance of this component (if any)
* @return Number    Resistance in ohms (can be infinity)
*/
ModelSwitch.prototype.getConstraintResistance = function() {
    return this.on ? 0 : Infinity;
}

/**
* Returns whether this component is symmetrical
* @return Boolean   Does the orientation of this component *not* make any difference?
*/
ModelSwitch.prototype.isSymmetric = function() {
    return false;
}

/**
* Create string representing component in serialized string
* @return String        String representing component (starting with corresponding letter registered in ModelComponent.serializingLetters) or false if component cannot be serialized
*/
ModelSwitch.prototype.serialize = function() {
    return this._on ? 'c' : 'o';
}

/**
* Which letter in serialized string correspondes to which child class of ModelComponent? Only lowercase is allowed, flippable components are represented by corresponding uppercase letter when flipped
*/
ModelComponent.serializingLetters['c'] = ModelSwitch;
ModelComponent.serializingLetters['o'] = ModelSwitch;

/**
* Create ModelSwitch instance from serialized string s
* @return ModelSwitch       Instance described by serialized string
*/
ModelSwitch.fromSerialized = function(
    circuit,    // ModelCircuit instance this instance should belong to
    s           // Serialized string describing component
) {
    var o = new ModelSwitch(circuit);
    o._on = (s == 'c');
    return o;
}

/**
* Set current simulation status (e. g. current)
* @return undefined
*/
ModelSwitch.prototype.setSimulationStatus = function(
    current         // Current through component (undefined → no simulation; Infinity → short circuit)
) {
    if (current === undefined)
        this._simulationToggled = false;
}
