/**
* An area of a circuit that can be simulated without considering further components
*/
function SimulationCircuitArea(
    width       // Width of hole circuit or area (neccessary to calculate ids for grid points
) {
    this._gridpointhash = new GridPointHasher(width);
    this.gridPoints = [];
    this.parts = [];
}

/** Arrays of grid point ids having same potential */
SimulationCircuitArea.prototype.gridPoints = [];

/** Array of relevant circuit parts having constrant voltage or constrant resistance */
SimulationCircuitArea.prototype.parts = [];

/** Object used to calculate grid point's ids */
SimulationCircuitArea.prototype._gridpointhash = null;

/**
* Calculate an ID of grid point gridpoint
* @return Number    ID of grid point
*/
SimulationCircuitArea.prototype.nameGridPoint = function(
    gridpoint       // Object linke { x: …, y: … }
) {
    return this._gridpointhash.calcHash(gridpoint);
}

/**
* Read relevant data for area containing ModelCircuitPart base from ModelCircuit instance circuit
* @return undefined
*/
SimulationCircuitArea.prototype.build = function(
    base,       // ModelCircuitPart instance of circuit that must be part of area
    circuit     // ModelCircuit instance containing all parts
) {
    const arrPartsUsed = [];
    const arrGridPointIdsUsed = [];
    let arrGridPointsLeft = [];
    const self = this;

    function addPart(part, forceInclude) {
        if (arrPartsUsed.indexOf(part) != -1)
            return;
        arrPartsUsed.push(part);
        const hasConstrantVoltage = (part.getConstraintVoltage() != 0);
        const resistance = part.getConstraintResistance();
        switch(true) {
            case !resistance:
                if (!hasConstrantVoltage)
                    break;
            case Number.isFinite(resistance):
            case hasConstrantVoltage:
                self.parts.push(part);
                break;
            case forceInclude:
                break;
            default:
                return;
        }
        const gridPoints = part.getAllConnectors();
        arrGridPointsLeft = arrGridPointsLeft.concat(gridPoints);
        const gridPointIds = gridPoints.map(gridpoint => self.nameGridPoint(gridpoint));
        if (resistance)
            gridPointIds.filter(id => self.gridPoints.reduce((prev, curr) => (prev && (curr.indexOf(id) == -1)), true)).forEach(id => self.gridPoints.push([ id ]));
        else {
            gridPointIds.forEach(id => self.gridPoints.push([ id ]));
            const sets = self.gridPoints.filter(set => gridPointIds.reduce((prev, curr) => (prev || (set.indexOf(curr) != -1)), false));
            self.gridPoints = self.gridPoints.filter(arr => (sets.indexOf(arr) == -1));
            self.gridPoints.push(sets.reduce((prev, curr) => { curr.forEach(id => { if (prev.indexOf(id) == -1) prev.push(id); }); return prev; }));
        }
    }

    addPart(base, true);
    while (arrGridPointsLeft.length) {
        const gridpoint = arrGridPointsLeft.pop();
        const id = this.nameGridPoint(gridpoint);
        if (arrGridPointIdsUsed.indexOf(id) != -1)
            continue;
        arrGridPointIdsUsed.push(id);
        circuit.getPartsConnectedTo(gridpoint.x, gridpoint.y).forEach(part => addPart(part));
    }

    // Find relevant meshes
    this.meshes = this.parts.filter(part => part.getConstraintVoltage()).map(part => circuit.getMeshes(part)).reduce((a, b) => a.concat(b), []);
}

/**
* Map connectors of part to indices of grid point set in this.gridPoints containing the corresponding grid point hash
* @return Array     Corresponding indices
*/
SimulationCircuitArea.prototype._getGridpointIndizes = function(
    part        // Item of this.parts
) {
//console.log({ part, connectors: part.getAllConnectors(), ids: part.getAllConnectors().map(gridpoint => this.nameGridPoint(gridpoint)), arrays: this.gridPoints });
    return part.getAllConnectors().map(gridpoint => this.nameGridPoint(gridpoint)).map(id => {
        for (let n = 1; n < this.gridPoints.length; ++n) {
            if (this.gridPoints[n].indexOf(id) != -1)
                return n;
        }
        return 0;
    });
}

/**
* Simulate circuit area and apply calculated currents and short circuits
* @return undefined
*/
SimulationCircuitArea.prototype.simulate = function(
    sheduleManager  // Responsible SimulationSheduleManager instance
) {
    // Find short circuits
    const ignoreParts = [];
    const ignorePartsStatus = [];
    this.meshes.forEach(mesh => {
        if (!mesh.isShortCircuit(true))
            return;
        const parts = mesh.getParts();
        parts.filter(part => (part.component instanceof ModelFuse)).forEach(part => {
            if (ignoreParts.indexOf(part) == -1) ignoreParts.push(part);
        });
        parts.filter(part => part.getConstraintVoltage()).forEach(part => { if (ignoreParts.indexOf(part) == -1) ignoreParts.push(part); });
    });
    // Ignore short circuited batteries
    let parts = this.parts;
    if (ignoreParts.length) {
        parts = parts.filter(part => (ignoreParts.indexOf(part) == -1));
        ignoreParts.forEach(part => {
            const gridpoints = this._getGridpointIndizes(part);
            if (gridpoints[0] == gridpoints[1])
                return;
            const sets = gridpoints.map(index => this.gridPoints[index]);
            this.gridPoints = this.gridPoints.filter((set, index) => (gridpoints.indexOf(index) == -1));
            this.gridPoints.push(sets.reduce((a, b) => a.concat(b), []));
        });
    }
    ignoreParts.forEach(part => ignorePartsStatus.push({ part, current: Infinity, voltage: 0 }));
    // Append indices of gridpoints to parts
    parts = parts.map(part => ({ part, gridpoints: this._getGridpointIndizes(part) }));
    // Find batteries
    const arrBatteries = parts.filter(({part}) => part.getConstraintVoltage());
    const constraintVoltages = arrBatteries.map(({part, gridpoints}) => {
        if (gridpoints[0] == gridpoints[1])
            return false;
        return { voltage: part.getConstraintVoltage(), from: gridpoints[0], to: gridpoints[1] };
    });
    // Add equations for batteries
    const vars = this.gridPoints.map(() => null).concat(parts);
    const equations = new LinearEquationSystem();
    constraintVoltages.forEach(({ voltage, from, to }) => {
        const equation = new LinearEquation();
        equation.setNumber(from, 1);
        equation.setNumber(to, -1);
        equation.setNumber(vars.length, voltage);
        equations.add(equation);
    });
    // Add equations for grid points and circuit parts
    const arrGridPointEquations = this.gridPoints.map(() => new LinearEquation());
    const varOffset = this.gridPoints.length;
    arrGridPointEquations.forEach(equation => equations.add(equation));
    parts.forEach(({ part, gridpoints }, index) => {
        // Ensure current 0A if short circuit
        if (gridpoints[0] == gridpoints[1]) {
            const equation = new LinearEquation();
            equation.setNumber(varOffset + index, 1);
            equations.add(equation);
            return;
        }
        // Insert to grid points' equation
        arrGridPointEquations[gridpoints[0]].setNumber(varOffset + index, +1);
        arrGridPointEquations[gridpoints[1]].setNumber(varOffset + index, -1);
        // Equation to ensure resistance?
        const resistance = part.getConstraintResistance();
        if (Number.isFinite(resistance)) {
            const equation = new LinearEquation();
            equation.setNumber(varOffset + index, resistance);
            equation.setNumber(gridpoints[0], -1);
            equation.setNumber(gridpoints[1], +1);
            equations.add(equation);
        }
    });
    // Solve linear equation system and interprete
    equations.solve();
//console.log(equations.toString());
    if (equations.isConflict())
        this._setSimulationStatus(ignorePartsStatus.concat(arrBatteries.map(({ part }) => ({ part, current: Infinity, voltage: 0 }))), sheduleManager);
    else
        this._setSimulationStatus(ignorePartsStatus.concat(parts.map(({ part, gridpoints }, index) => ({
            part,
            current: equations.getSolution(varOffset + index),
            voltage: equations.calcDifference(gridpoints[1], gridpoints[0])
        }))), sheduleManager);
}

/**
* Set current of circuit parts to given values (and current of other circuit parts to 0)
* @return undefined
*/
SimulationCircuitArea.prototype._setSimulationStatus = function(
    nonZero,        // Array of items like { part: "item of this.part", current: "calculated current" } for all items of this.part not having current 0
    sheduleManager  // Responsible SimulationSheduleManager instance
) {
    const all = nonZero.slice();
    const used = nonZero.map(({part}) => part);
    this.parts.filter(part => (used.indexOf(part) == -1)).forEach(part => all.push({ part, current: 0, voltage: 0 }));
    all.forEach(({ part, current, voltage }) => part.setSimulationStatus(current, voltage, sheduleManager));
    this.parts = all.map(({part}) => part);
}
