/**
* Class pooling time-displaced actions while simulating having same delay
*/
function SimulationSheduleAction(
    manager,        // Responsible SimulationSheduleManager instance
    delay           // Delay for execution in milli seconds
) {
    Object.defineProperty(this, 'manager', { enumerable: true, value: manager, writeable: false });
    Object.defineProperty(this, 'delay', { enumerable: true, value: delay, writeable: false });
    this._actions = [];
}

/** Responsible SimulationSheduleManager instance */
SimulationSheduleAction.prototype.manager = null;

/** Delay for execution in milli seconds */
SimulationSheduleAction.prototype.delay = 0;

/** Id of timeout */
SimulationSheduleAction.prototype._timeoutId = undefined;

/** Planed actions as objects like { id: 1, fkt: function() { do_something(); } } */
SimulationSheduleAction.prototype._actions = [];

/**
* Start timeout
* @return undefined
*/
SimulationSheduleAction.prototype.start = function() {
    this._timeoutId = window.setTimeout(() => this._execute(), this.delay);
}

/**
* Abort all actions
* @return undefined
*/
SimulationSheduleAction.prototype.stop = function() {
    if (this._timeoutId !== undefined) {
        window.clearTimeout(this._timeoutId);
        delete this._timeoutId;
    }
    this.manager.removeAction(this);
}

/**
* Remove time-displaced action with given id
* @return Boolean       Successfully removed?
*/
SimulationSheduleAction.prototype.remove = function(
    needle      // Id of action to be removed
) {
    let removed = false;
    this._actions = this._actions.filter(({ id, fkt }) => { if (id == needle) { removed = true; return false; } else return true; });
    if (!this._actions.length)
        this.stop();
    return removed;
}

/**
* Add time-displaced action
* @return Number        Created id for action
*/
SimulationSheduleAction.prototype.add = function(
    fkt         // Function to call after delay
) {
    const id = this.manager.createId();
    this._actions.push({ id, fkt });
    return id;
}

/**
* Remove this instance from responsible SimulationSheduleManager instance and execute all time-displaced actions
* @return undefined
*/
SimulationSheduleAction.prototype._execute = function() {
    this.manager.removeAction(this);
    this._actions.forEach(({ fkt }) => fkt.call(null));
}
