/**
* Class for pooling time-displaced actions while simulating
*/
function SimulationSheduleManager() {
    this._waiting = [];
    this._preplaned = [];
}

/** All SimulationSheduleAction instances waiting to fire */
SimulationSheduleManager.prototype._waiting = [];

/** All SimulationSheduleAction instances without startet timeout */
SimulationSheduleManager.prototype._preplaned = [];

/** Highest id assigned */
SimulationSheduleManager.prototype._lastId = 0;

/**
* Create an id
* @return Number        Created id
*/
SimulationSheduleManager.prototype.createId = function() {
    return ++this._lastId;
}

/**
* Remove (already stopped or never started!) SimulationSheduleAction instance
* @return undefined
*/
SimulationSheduleManager.prototype.removeAction = function(
    action      // SimulationSheduleAction instance to remove
) {
    this._waiting = this._waiting.filter(item => (item != action));
    this._preplaned = this._preplaned.filter(item => (item != action));
}

/**
* Add time-displaced action
* @return Number        Created id for action
*/
SimulationSheduleManager.prototype.add = function(
    fkt,        // Function to call after delay
    delay       // Delay for execution in milli seconds
) {
    const actions = this._preplaned.filter(action => (action.delay == delay));
    const action = actions.length ? actions[0] : new SimulationSheduleAction(this, delay);
    if (!actions.length)
        this._preplaned.push(action);
    return action.add(fkt);
}

/**
* Remove time-displaced action with given id
* @return Boolean       Successfully removed?
*/
SimulationSheduleManager.prototype.remove = function(
    needle      // Id of action to be removed
) {
    return this._waiting.reduce((prev, curr) => (prev || curr.remove(needle)), this._preplaned.reduce((prev, curr) => (prev || curr.remove(needle)), false));
}

/**
* Start timeout of all SimulationSheduleAction instances without startet timeout
* @return undefined
*/
SimulationSheduleManager.prototype.start = function() {
    while (this._preplaned.length) {
        const action = this._preplaned.pop();
        this._waiting.push(action);
        action.start();
    }
}

/**
* Abort all actions
* @return undefined
*/
SimulationSheduleManager.prototype.stop = function() {
    this._waiting.forEach(action => action.stop());
}
