/**
* Class used to create an actual SVG graphic
*/
function ViewCircuit(
    document,       // Document instance in which SVG should be created
    dimensions      // ViewCircuitDimensions instance providing messurements
) {
    this.node = document.createElementNS(ViewCircuit.ns, 'svg');
    this.node.setAttribute('width', dimensions.localWidth + 'mm');
    this.node.setAttribute('height', dimensions.localHeight + 'mm');
    this.node.setAttribute('viewBox', [dimensions.minLocalX, dimensions.minLocalY, dimensions.localWidth, dimensions.localHeight].join(' '));
    this._layers = this._layers.slice();
    const defs = document.createElementNS(ViewCircuit.ns, 'defs');
    this._styleNode = document.createElementNS(ViewCircuit.ns, 'style');
    defs.appendChild(this._styleNode);
    this.node.appendChild(defs);
}

/** SVG Namespace */
ViewCircuit.ns = 'htt' + 'p://w' + 'ww.w3.org/2000/svg';

/** Created SVG node */
ViewCircuit.prototype.node = null;

/** Node for global SVG CSS styles */
ViewCircuit.prototype._styleNode = null;

/** Node for metadata */
ViewCircuit.prototype._metadataNode = null;

/** All layers */
ViewCircuit.prototype._layers = [null, null, null, null];

/** Constants for possible layers */
ViewCircuit.LAYER_BACKGROUND = 0;       // Layer for background elements like highlighters
ViewCircuit.LAYER_GRID = 1;             // Layer for grid lines
ViewCircuit.LAYER_COMPONENTS = 2;       // Layer for components
ViewCircuit.LAYER_INTERACT = 3;         // Topmost layer for clickable elements

/**
* Add class name and CSS to newly created layer
* @return undefined
*/
ViewCircuit.prototype._furnishLayer = function(
    layer       // Layer by ViewCircuit.LAYER_* constant
) {
    switch (layer) {
        case ViewCircuit.LAYER_COMPONENTS:
            this._styleNode.textContent += '.components * {fill:none;stroke:#000;stroke-width:.4;stroke-linecap:square}';
            this._layers[layer].node.setAttribute('class', 'components');
            break;
        case ViewCircuit.LAYER_GRID:
            this._styleNode.textContent += '.grid * {stroke:#888a85;stroke-width:0.2;stroke-linecap:round;stroke-linejoin:round;stroke-miterlimit:4;stroke-dasharray:0.6, 0.6}';
            this._layers[layer].node.setAttribute('class', 'grid');
            break;
        case ViewCircuit.LAYER_BACKGROUND:
            this._styleNode.textContent += '.highlight * {fill:none} .highlight .good {fill:#8ae234} .highlight .ok {fill:#fce94f} .highlight .bad {fill:#d3d7cf}';
            this._layers[layer].node.setAttribute('class', 'highlight');
            break;
        case ViewCircuit.LAYER_INTERACT:
            this._styleNode.textContent += '.click * {opacity:0;fill:#fff;outline:0 none;stroke:none;cursor:pointer}';
            this._layers[layer].node.setAttribute('class', 'click');
            break;
        /*!*/
    }
}

/**
* Get ViewCircuitGroup instance corresponding to specified layer (create first if neccessary)
* @return ViewCircuitGroup      ViewCircuitGroup instance corresponding to layer
*/
ViewCircuit.prototype.getLayer = function(
    layer       // Layer by ViewCircuit.LAYER_* constant
) {
    var index = Math.max(0, Math.min(this._layers.length - 1, Math.round(layer)));
    if (!(this._layers[index] instanceof ViewCircuitGroup)) {
        var g = document.createElementNS(ViewCircuit.ns, 'g');
        this.node.insertBefore(g, this._layers.slice(index + 1).reduce(function(prev, next) { return prev || (next && next.node); }, null));
        this._layers[index] = new ViewCircuitGroup(g);
        this._furnishLayer(index);
    }
    return this._layers[index];
}

/**
* Create a new node in metadata area
* @return Element       Created element
*/
ViewCircuit.prototype.addMetadata = function(
    namespaceURI,       // Namespace of element to add
    nodeName            // Name of node to add
) {
    const doc = this.node.ownerDocument;
    if (!this._metadataNode) {
        this._metadataNode = doc.createElementNS(ViewCircuit.ns, 'metadata');
        this.node.insertBefore(this._metadataNode, this.node.firstChild);
    }
    const node = doc.createElementNS(namespaceURI, nodeName);
    this._metadataNode.appendChild(node);
    return node;
}
