/**
* Class for creating a SVG showing the present state of ModelCircuit instance circuit
*/
function ViewCircuitBuilder(
    circuit,        // ModelCircuit instance containing data of circuit to render
    document,       // Document instance in which SVG should be created
    scale,          // Length of grid cell in SVG local units
    minWidth,       // Optional: Minimal image width in SVG local units
    minHeight,      // Optional: Minimal image height in SVG local units
    offsetX,        // Optional: Free space left in grid units
    offsetY         // Optional: Free space top in grid units
) {
    this._circuit = circuit;
    this._contentDimensions = circuit.getUsedGridArea();
    this._imageDimensions = new ViewCircuitDimensions(
        this._contentDimensions.left - 0.5 - (offsetX ? offsetX : 0),
        this._contentDimensions.top - 0.5 - (offsetY ? offsetY : 0),
        Math.max(this._contentDimensions.right, Number.isFinite(minWidth / scale) ? (minWidth / scale + this._contentDimensions.left - 1) : Number.NEGATIVE_INFINITY) + 0.5 - (offsetX ? offsetX : 0),
        Math.max(this._contentDimensions.bottom, Number.isFinite(minHeight / scale) ? (minHeight / scale + this._contentDimensions.top - 1) : Number.NEGATIVE_INFINITY) + 0.5 - (offsetY ? offsetY : 0),
        scale
    );
    this.view = new ViewCircuit(document, this._imageDimensions);
}

/** Used ViewCircuit instance */
ViewCircuitBuilder.prototype.view = null;

/** ModelCircuit instance representating data to render */
ViewCircuitBuilder.prototype._circuit = null;

/** Dimensions of content */
ViewCircuitBuilder.prototype._contentDimensions = null;

/** Dimensions of image */
ViewCircuitBuilder.prototype._imageDimensions = null;

/** Created SVG node */
Object.defineProperty(ViewCircuitBuilder.prototype, 'node', {
    enumerable: true,
    get() { return (this.view instanceof ViewCircuit) ? this.view.node : null; }
});

/** Default functions to use for rendering circuit parts */
ViewCircuitBuilder.defaultRenderingFunctions = [renderConnection, renderLightbulb, renderSwitch, renderMonocell, renderFuse, renderMeter];

/**
* Add serialized circuit data to SVG image
* @return undefined
*/
ViewCircuitBuilder.prototype.addData = function() {
    this.view.addMetadata('https://www.roith-unterrichtet.de', 'circuitdata').textContent = this._circuit.serialize();
    /*!*/
}

/**
* Add all circuit parts to SVG image
* @return undefined
*/
ViewCircuitBuilder.prototype.addCircuitParts = function(
    arrRenderingFkts        // Optional: Functions to use for rendering
) {
    var view = this.view;
    var dimensions = this._imageDimensions;
    var fktMap = (arrRenderingFkts || ViewCircuitBuilder.defaultRenderingFunctions).reduce(function(a, b) {
        return function(view, dimensions, circuitpart) { return a.apply(null, arguments) || b.apply(null, arguments); };
    });
    this._circuit.map(function(circuitpart) { return fktMap(view, dimensions, circuitpart); });
}

/**
* Add dots for open and multiple connection points (only call after addCircuitParts!)
* @return undefined
*/
ViewCircuitBuilder.prototype.addDots = function() {
    var x, y, count;
    for (y = this._contentDimensions.top; y <= this._contentDimensions.bottom; ++y) {
        for (x = this._contentDimensions.left; x <= this._contentDimensions.right; ++x) {
            count = this._circuit.getPartsConnectedTo(x, y).length;
            if (count && (count != 2))
                renderDot(this.view, this._imageDimensions, x, y, count == 1);
        }
    }
}

/**
* Add visible grid
* @return undefined
*/
ViewCircuitBuilder.prototype.addGrid = function(
    noCroppedLines      // Optional: Draw only lines between visible grid points?
) {
    var gridLeft = Math.ceil(this._imageDimensions.minInternalX);
    var gridRight = Math.floor(this._imageDimensions.maxInternalX);
    var gridTop = Math.ceil(this._imageDimensions.minInternalY);
    var gridBottom = Math.floor(this._imageDimensions.maxInternalY);
    var left = noCroppedLines ? this._imageDimensions.internal2localX(gridLeft) : this._imageDimensions.minLocalX;
    var right = noCroppedLines ? this._imageDimensions.internal2localX(gridRight) : this._imageDimensions.maxLocalX;
    var top = noCroppedLines ? this._imageDimensions.internal2localY(gridTop) : this._imageDimensions.minLocalY;
    var bottom = noCroppedLines ? this._imageDimensions.internal2localY(gridBottom) : this._imageDimensions.maxLocalY;
    var n, pos;
    var layer = this.view.getLayer(ViewCircuit.LAYER_GRID);
    for (n = gridLeft; n <= gridRight; ++n) {
        pos = this._imageDimensions.internal2localX(n);
        layer.addElement('line', { x1: pos, y1: top, x2: pos, y2: bottom });
    }
    for (n = gridTop; n <= gridBottom; ++n) {
        pos = this._imageDimensions.internal2localY(n);
        layer.addElement('line', { x1: left, y1: pos, x2: right, y2: pos });
    }
}

/**
* Add selectors for all grid places
* @return undefined
*/
ViewCircuitBuilder.prototype.addSelectors = function(
    controller      // ControllerCircuit instance to notify when selected and to get information on clickable state from
) {
    var x, y, isHorizontal;
    var minY = Math.floor(this._imageDimensions.minInternalY) * 2 + 1;
    var max2Y = Math.floor(this._imageDimensions.maxInternalY) * 2 + 2;
    var minX = Math.floor(this._imageDimensions.minInternalX);
    var maxX = Math.floor(this._imageDimensions.maxInternalX);
    for (y = minY; y < max2Y; ++y) {
        isHorizontal = (y % 2) ? false : true;
        for (x = minX + (isHorizontal ? 0 : 1); x <= maxX; ++x)
            new ViewPlaceSelector(controller, this.view, this._imageDimensions, x, Math.floor(y / 2), isHorizontal);
    }
}
