/**
* Hoverable and clickable selector for a certain place within the circuit
*/
function ViewPlaceSelector(
    controller,     // ControllerCircuit instance to notify when selected and to get information on clickable state from
    view,           // SVG (ViewCircuit instance) used for rendering
    dimensions,     // ViewCircuitDimensions instance providing messurements
    x,              // Horizontal internal coordinate of grid point at which to start
    y,              // Vertical internal coordinate of grid point at which to start
    isHorizontal    // Is it a selector for a horizontal place?
) {
    this.controller = controller;
    this.x = x;
    this.y = y;
    if (isHorizontal)
        this.isHorizontal = true;
    this._nodeHighlighter = this._createPath(view.getLayer(ViewCircuit.LAYER_BACKGROUND), dimensions);
    this._nodeCatcher = this._createPath(view.getLayer(ViewCircuit.LAYER_INTERACT), dimensions);
    this._nodeCatcher.setAttribute('tabindex', 0);
    var self = this;
    this._nodeCatcher.addEventListener('focus', function() { self.updateHighlighter(); });
    this._nodeCatcher.addEventListener('blur', function() { self.updateHighlighter(); });
    this._nodeCatcher.addEventListener('mouseenter', function() { self.isHovered = true; self.updateHighlighter(); });
    this._nodeCatcher.addEventListener('mouseleave', function() { self.isHovered = false; self.updateHighlighter(); });
    this._nodeCatcher.addEventListener('click', function(ev) { self.procSelected(ev.shiftKey, ev.ctrlKey, ev.altKey); });
    this._nodeCatcher.addEventListener('keypress', function(ev) {
        switch (ev.key) {
            case 'Enter':
            case ' ':
                self.procSelected(ev.shiftKey, ev.ctrlKey, ev.altKey);
                break;
        }
    });
    /*!*/
}

/** ControllerCircuit instance to notify when selected and to get information on clickable state from */
ViewPlaceSelector.prototype.controller = null;

/** Horizontal internal coordinate of grid point at which to start */
ViewPlaceSelector.prototype.x = 0;

/** Vertical internal coordinate of grid point at which to start */
ViewPlaceSelector.prototype.y = 0;

/** Is it a selector for a horizontal place? */
ViewPlaceSelector.prototype.isHorizontal = 0;

/** Shape used for highlighting */
ViewPlaceSelector.prototype._nodeHighlighter = null;

/** Shape used for catching mouse events */
ViewPlaceSelector.prototype._nodeCatcher = null;

/** Is mouse over this place? */
ViewPlaceSelector.prototype.isHovered = false;

/**
* Update state of highlighter
* @return undefined
*/
ViewPlaceSelector.prototype.updateHighlighter = function() {
    if (this.isHovered ||(this._nodeCatcher == document.activeElement))
        this._nodeHighlighter.setAttribute('class', this.controller.getPlaceClass(this));
    else
        this._nodeHighlighter.removeAttribute('class');
}

/**
* Create and add a diamond shaped path
* @return SVGPathElement        Created path element
*/
ViewPlaceSelector.prototype._createPath = function(
    layer,              // ViewCircuitGroup instance to add path node to
    dimensions          // ViewCircuitDimensions instance providing messurements
) {
    return layer.addElement('path', {
        d: 'm' + dimensions.internal2localX(this.x) + ' ' + dimensions.internal2localY(this.y) + 'l' + [
            [0.25, 0.25],
            this.isHorizontal ? [0.5, 0] : [0, 0.5],
            this.isHorizontal ? [0.25, -0.25] : [-0.25, 0.25],
            [-0.25, -0.25],
            this.isHorizontal ? [-0.5, 0] : [0, -0.5]
        ].map(function(v) {
            return dimensions.internal2local(v[0]) + ' ' + dimensions.internal2local(v[1]);
        }).join(' ') + 'z'
    });
}

/**
* Tell controller that place has been selected
* @return undefined
*/
ViewPlaceSelector.prototype.procSelected = function(
    shiftKey,   // Is shift pressed?
    ctrlKey,    // Is ctrl pressed?
    altKey      // Is alt pressed?
) {
    this.controller.procSelected(this, shiftKey, ctrlKey, altKey);
}
