/**
* Draw monocell in SVG
* @return boolean       Did this function render circuitpart?
*/
function renderMonocell(
    view,           // SVG (ViewCircuit instance) used for rendering
    dimensions,     // ViewCircuitDimensions instance providing messurements
    circuitpart     // ModelCircuitPart instance to eventually render
) {
    if (!(circuitpart.component instanceof ModelMonocell))
        return false;
    var group = view.getLayer(ViewCircuit.LAYER_COMPONENTS).addGroup();

    var place = circuitpart.getUsedGridArea();
    var minX = !circuitpart.isFlipped ? place.right : place.left;
    var maxX = !circuitpart.isFlipped ? place.left : place.right;
    var minY = !circuitpart.isFlipped ? place.bottom : place.top;
    var maxY = !circuitpart.isFlipped ? place.top : place.bottom;

    var abstand = (1 - renderMonocell.cellLength) / 2;
    var cx1 = (1 - abstand) * minX + abstand * maxX;
    var cy1 = (1 - abstand) * minY + abstand * maxY;
    var cx2 = abstand * minX + (1 - abstand) * maxX;
    var cy2 = abstand * minY + (1 - abstand) * maxY;

    function drawLine(x1, y1, x2, y2) {
        group.addElement('line', {
            x1: dimensions.internal2localX(x1),
            y1: dimensions.internal2localY(y1),
            x2: dimensions.internal2localX(x2),
            y2: dimensions.internal2localY(y2)
        });
    }

    function drawSign(x, y, boolPlus, boolLeft, boolBottom) {
        x += boolLeft ? (-renderMonocell.textPadding - renderMonocell.barLength) : renderMonocell.textPadding;
        y += boolBottom ? (renderMonocell.textPadding + renderMonocell.barLength) : (-renderMonocell.textPadding);
        drawLine(x, y - renderMonocell.barLength / 2, x + renderMonocell.barLength, y - renderMonocell.barLength / 2);
        if (boolPlus)
            drawLine(x + renderMonocell.barLength / 2, y, x + renderMonocell.barLength / 2, y - renderMonocell.barLength);
    }

    drawLine(minX, minY, cx1, cy1);
    // Plus pole
    drawLine(
        cx1 - renderMonocell.plusLength * (maxY - minY) / 2,
        cy1 + renderMonocell.plusLength * (maxX - minX) / 2,
        cx1 + renderMonocell.plusLength * (maxY - minY) / 2,
        cy1 - renderMonocell.plusLength * (maxX - minX) / 2
    );
    if (!circuitpart.component.shortCircuited)
        drawSign(cx1, cy1, true, minX < maxX, maxY < minY);
    // Minus pole
    drawLine(
        cx2 - renderMonocell.minusLength * (maxY - minY) / 2,
        cy2 + renderMonocell.minusLength * (maxX - minX) / 2,
        cx2 + renderMonocell.minusLength * (maxY - minY) / 2,
        cy2 - renderMonocell.minusLength * (maxX - minX) / 2
    );
    if (!circuitpart.component.shortCircuited)
        drawSign(cx2, cy2, false, minX > maxX, maxY > minY);
    drawLine(cx2, cy2, maxX, maxY);
    if (circuitpart.component.shortCircuited) {
        if (place.left == place.right) {
            renderMonocell.drawFlash(
                group,
                dimensions.internal2localX(place.left + 0.2 * abstand),
                dimensions.internal2localY((place.top + place.bottom) / 2),
                dimensions.internal2localY(abstand) - dimensions.internal2localY(0)
            );
        } else {
            renderMonocell.drawFlash(
                group,
                dimensions.internal2localX((Math.max(cx1, cx2) + place.right) / 2 - 0.2 * abstand),
                dimensions.internal2localY((minY + maxY) / 2),
                dimensions.internal2localY(abstand) - dimensions.internal2localY(0)
            );
            /*!*/
        }
    }
    return true;
}

/** Distance between lines in internal units */
renderMonocell.cellLength = 0.1;

/** Length of plus pole in internal units */
renderMonocell.plusLength = 0.4;

/** Length of minus pole in internal units */
renderMonocell.minusLength = 0.2;

/** Length of bars of plus and minus sign in internal units */
renderMonocell.barLength = 0.1;

/** Distance between circuit symbol and plus or minus sign in internal units */
renderMonocell.textPadding = 0.07;

/**
* Adds a flash sign to viewGroup (width is 0.4 * height
* @return undefined
*/
renderMonocell.drawFlash = function(
    viewGroup,      // ViewCircuitGroup instance to draw to
    x,              // Position of left outline of flash in local units
    y,              // Distance from top of image to center of flash in local units
    height          // Height of flash in local units
) {
    const pathparts = [];
    pathparts.push('m' + (x + 0.11 * height) + ' ' + (y - 0.5 * height));
    pathparts.push('h' + (0.26 * height));
    pathparts.push('l' + (-0.16 * height) + ' ' + (0.33 * height));
    pathparts.push('h' + (0.2 * height));
    pathparts.push('l' + (-0.40 * height) + ' ' + (0.67 * height));
    pathparts.push((0.15 * height) + ' ' + (-0.57 * height));
    pathparts.push('h' + (-0.17 * height));
    pathparts.push('z');
    viewGroup.addElement('path', { d: pathparts.join(' '), style: 'fill:#fce94f;stroke:#c4a000' });
}
